import os
import sys
import subprocess
import re

def install_requirements():
    required_packages = [
        "gradio",
        "torch",
        "transformers",
        "diffusers",
        "Pillow",
        "numpy",
        "huggingface_hub"
    ]
    for package in required_packages:
        try:
            __import__(package)
        except ImportError:
            print(f"Installing missing package: {package}")
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])

install_requirements()

import gradio as gr
import torch
from janus.janusflow.models import MultiModalityCausalLM, VLChatProcessor
from PIL import Image  # Only import Image
from diffusers import DiffusionPipeline
import numpy as np
import random
from huggingface_hub import snapshot_download
import shutil

# Ensure GPU
if not torch.cuda.is_available():
    raise EnvironmentError("GPU is required for this script.")

cuda_device = 'cuda'
torch.backends.cuda.matmul.allow_tf32 = True
torch.backends.cudnn.benchmark = True
torch.backends.cudnn.deterministic = False
torch.cuda.empty_cache()

cache_dir = os.path.expanduser("~/.cache/huggingface/hub")
UPLOAD_FOLDER = "uploaded_images"
os.makedirs(UPLOAD_FOLDER, exist_ok=True)

def model_exists(model_name):
    model_path = os.path.join(cache_dir, f"models--{model_name.replace('/', '--')}")
    return os.path.exists(model_path)

def load_janusflow():
    global vl_gpt, vl_chat_processor, tokenizer
    if 'vl_gpt' not in globals():
        print("🔹 Loading JanusFlow Model...")
        model_name = "deepseek-ai/JanusFlow-1.3B"
        if not model_exists(model_name):
            snapshot_download(repo_id=model_name, cache_dir=cache_dir)
        vl_chat_processor = VLChatProcessor.from_pretrained(model_name)
        tokenizer = vl_chat_processor.tokenizer
        vl_gpt = MultiModalityCausalLM.from_pretrained(
            model_name, torch_dtype=torch.float16
        ).to(cuda_device).eval()
        print("✅ JanusFlow Model Loaded.")

def load_flux():
    global flux_pipeline
    if 'flux_pipeline' not in globals():
        print("🔹 Loading FLUX Model...")
        model_name = "black-forest-labs/FLUX.1-schnell"
        if not model_exists(model_name):
            snapshot_download(repo_id=model_name, cache_dir=cache_dir)
        torch.cuda.empty_cache()
        torch.cuda.ipc_collect()
        flux_pipeline = DiffusionPipeline.from_pretrained(
            model_name,
            torch_dtype=torch.float16,
            device_map="balanced"
        )
        try:
            if hasattr(flux_pipeline, "enable_xformers_memory_efficient_attention"):
                flux_pipeline.enable_xformers_memory_efficient_attention()
        except Exception as e:
            print(f"Warning: {e}")
        print("✅ FLUX Model Loaded.")

def safe_seed(seed):
    return int(seed) % (2**32)

def resize_large_image(image):
    max_size = 1024
    if isinstance(image, Image.Image):
        if max(image.size) > max_size:
            # Use LANCZOS instead of deprecated ANTIALIAS
            image.thumbnail((max_size, max_size), Image.LANCZOS)
    return image

def copy_uploaded_files(filepaths):
    saved_paths = []
    for fpath in filepaths:
        base = os.path.basename(fpath)
        dst = os.path.join(UPLOAD_FOLDER, base)
        shutil.copy(fpath, dst)  # Overwrite existing if same name
        saved_paths.append(dst)
    return saved_paths

@torch.inference_mode()
def multimodal_understanding(all_paths, selected_index, question, seed, top_p, temperature):
    load_janusflow()
    torch.cuda.empty_cache()

    if not all_paths:
        return "No images uploaded."
    if selected_index < 0 or selected_index >= len(all_paths):
        return "No valid image selected."

    image_path = all_paths[selected_index]
    if not os.path.exists(image_path):
        return "No valid image path."

    pil_img = Image.open(image_path).convert("RGB")
    pil_img = resize_large_image(pil_img)
    np_img = np.array(pil_img)

    seed = safe_seed(seed)
    torch.manual_seed(seed)
    np.random.seed(seed)
    torch.cuda.manual_seed(seed)

    conversation = [
        {"role": "User", "content": f"<image_placeholder>\n{question}", "images": [np_img]},
        {"role": "Assistant", "content": ""},
    ]

    prepare_inputs = vl_chat_processor(
        conversations=conversation,
        images=[pil_img],
        force_batchify=True
    ).to(cuda_device, dtype=torch.float16)

    inputs_embeds = vl_gpt.prepare_inputs_embeds(**prepare_inputs)
    outputs = vl_gpt.language_model.generate(
        inputs_embeds=inputs_embeds,
        attention_mask=prepare_inputs.attention_mask,
        pad_token_id=vl_chat_processor.tokenizer.eos_token_id,
        max_new_tokens=400,
        do_sample=(temperature != 0),
        temperature=temperature,
        top_p=top_p
    )

    raw_response = vl_chat_processor.tokenizer.decode(outputs[0].cpu().tolist(), skip_special_tokens=True)

    # Insert a space between a lower letter and an uppercase letter (brownThe -> brown The)
    step1 = re.sub(r'([a-z])([A-Z])', r'\1 \2', raw_response)

    # Replace any other disallowed chars with a space
    # Allowed: letters, commas, periods, whitespace
    step2 = re.sub(r'[^A-Za-z,.\s]', ' ', step1)

    # Collapse multiple spaces into one, then strip
    filtered = re.sub(r'\s+', ' ', step2).strip()

    txt_path = os.path.splitext(image_path)[0] + ".txt"
    try:
        with open(txt_path, "w", encoding="utf-8") as f:
            f.write(filtered)
    except Exception as e:
        print(f"Warning: Could not write .txt: {e}")

    return filtered

@torch.inference_mode()
def generate_image(prompt, seed, cfg_weight, steps, randomize):
    load_flux()
    torch.cuda.empty_cache()
    torch.cuda.ipc_collect()

    if randomize:
        seed = random.randint(0, 2**32 - 1)
    seed = safe_seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    np.random.seed(seed)

    with torch.amp.autocast(device_type="cuda", dtype=torch.float16):
        result = flux_pipeline(
            prompt=prompt,
            guidance_scale=cfg_weight,
            num_inference_steps=steps,
            width=1024,
            height=1024
        )
        image = result.images[0]

    image.save("generated_image.png")
    del result
    torch.cuda.empty_cache()
    torch.cuda.ipc_collect()
    return image

def randomize_seed_button():
    return random.randint(0, 2**32 - 1)

def handle_file_change(paths, old_paths):
    if not paths:
        return old_paths, gr.update(value=None, visible=False, interactive=False), 0, gr.update(open=True)
    new_paths = copy_uploaded_files(paths)
    combined = old_paths + new_paths
    return combined, gr.update(value=combined, visible=True, interactive=True), 0, gr.update(open=False)

def on_gallery_select(evt: gr.SelectData, all_paths):
    return evt.index, gr.update()

def open_folder():
    folder_path = os.path.abspath(UPLOAD_FOLDER)
    if sys.platform.startswith("win"):
        os.startfile(folder_path)
    elif sys.platform.startswith("darwin"):
        subprocess.Popen(["open", folder_path])
    else:
        subprocess.Popen(["xdg-open", folder_path])
    return f"Opened folder: {folder_path}"

with gr.Blocks(title="Janus & FLUX Model") as demo:
    with gr.Tabs():
        with gr.Tab("Multimodal Understanding"):
            gr.Markdown("## Multimodal Understanding")

            with gr.Accordion("Upload Images (up to 20)", open=True) as upload_accordion:
                file_input = gr.File(
                    file_count="multiple",
                    type="filepath",
                    interactive=True,
                    label="Select or drag multiple images"
                )

            all_paths_state = gr.State([])
            selected_index_state = gr.State(0)

            gallery = gr.Gallery(
                label="Images",
                columns=1,
                height=512,
                object_fit="contain",
                visible=False,
                interactive=False
            )

            file_input.change(
                fn=handle_file_change,
                inputs=[file_input, all_paths_state],
                outputs=[all_paths_state, gallery, selected_index_state, upload_accordion]
            )

            gallery.select(
                fn=on_gallery_select,
                inputs=[all_paths_state],
                outputs=[selected_index_state, upload_accordion]
            )

            with gr.Row():
                question_input = gr.Textbox(label="Chat with Image", placeholder="Ask a question...")
                seed_input = gr.Number(label="Seed", value=42, precision=0)
                top_p = gr.Slider(0, 1, value=0.2, label="Top-p")
                temperature = gr.Slider(0, 1, value=0.2, label="Temperature")

            response_output = gr.Textbox(label="Response", lines=4)
            chat_button = gr.Button("💬 Chat")

            description_buttons = [
                ("👤 Person Only", "Only describe the person. Ignore all background, setting, and lighting."),
                ("📝 Brief", "Brief length description."),
                ("📘 Medium", "Medium length description."),
                ("📗 Detailed", "Detailed length description."),
                ("📚 Highly Detailed", "Highly detailed description with intricate details and unique aspects.")
            ]

            with gr.Row():
                for btn_text, preset_query in description_buttons:
                    gr.Button(btn_text).click(
                        fn=multimodal_understanding,
                        inputs=[
                            all_paths_state,
                            selected_index_state,
                            gr.Textbox(value=preset_query, visible=False),
                            seed_input,
                            top_p,
                            temperature
                        ],
                        outputs=response_output
                    )

            with gr.Row():
                open_btn = gr.Button("Open Uploaded Folder")
                open_folder_status = gr.Textbox(value="", visible=False)

            open_btn.click(fn=open_folder, outputs=open_folder_status)

            chat_button.click(
                fn=multimodal_understanding,
                inputs=[all_paths_state, selected_index_state, question_input, seed_input, top_p, temperature],
                outputs=response_output
            )

        with gr.Tab("Text-to-Image Generation"):
            gr.Markdown("## Text-to-Image Generation")
            cfg_weight_input = gr.Slider(1, 10, value=3.6, label="CFG Weight")
            step_input = gr.Slider(1, 50, value=30, label="Number of Inference Steps")
            prompt_input = gr.Textbox(label="Prompt", lines=3)
            seed_input_tti = gr.Number(label="Seed (Optional)", value=123456789, precision=0)

            with gr.Row():
                randomize_seed_btn = gr.Button("Randomize Seed")
                generate_button = gr.Button("Generate Images")

            image_output = gr.Image(label="Generated Image", height=512, width=512)

            randomize_seed_btn.click(fn=randomize_seed_button, outputs=seed_input_tti)
            generate_button.click(
                fn=generate_image,
                inputs=[prompt_input, seed_input_tti, cfg_weight_input, step_input, gr.State(False)],
                outputs=image_output
            )

demo.launch(share=True)
